// AnalogIO.h : main header file for the AnalogIO DLL
//

#if !defined(AFX_AnalogIO_H__65E6C3BC_87B9_11D3_8080_00105A1E5D53__INCLUDED_)
#define AFX_AnalogIO_H__65E6C3BC_87B9_11D3_8080_00105A1E5D53__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#ifndef __AFXWIN_H__
#error include 'stdafx.h' before including this file for PCH
#endif

#include "Resource.h"           // main symbols
#include "I2CMgr.h"
#include "I2CDeviceClass.h"
#include "Module.h"
#include "Xpert/Win32Fix.h"

#undef DECL
#ifdef ENGINE_DLL
#define DECL _declspec(dllexport)
#else
#define DECL _declspec(dllimport)
#endif

const int NUM_ANALOGCHANNELS = 16;
const CTimeSpan RecalInterval = 600;

// CAnalogChannel is a class that seemlessly deals with converting a 1-8 channel
// being passed in to this module which uses 0-7 numbering.
//
class DECL CAnalogChannel
{
public:
   CAnalogChannel(int nChannel)                   { Set(nChannel);                                      };
   CAnalogChannel()                               { Channel = 0; Valid = false;                         };
   CAnalogChannel(TProperty& Property)            { Set(Property.AsInteger());                          }; 
   CAnalogChannel(TValue& Value)                  { Set(Value.AsInteger());                             };
   CAnalogChannel(const CAnalogChannel& dc)       { Channel = dc.Channel; Valid=dc.Valid;               };
   CAnalogChannel& operator=(CAnalogChannel dc)   { Channel = dc.Channel; Valid=dc.Valid; return *this; };
   CAnalogChannel& operator=(int nChannel)        { Set(nChannel); return *this;                        };
   CAnalogChannel& operator=(TProperty& Property) { Set(Property.AsInteger()); return *this;            };
   CAnalogChannel& operator=(TValue& Value)       { Set(Value.AsInteger()); return *this;               };
   operator BYTE() const                          { return Channel;                                     };
   bool IsValid() const                           { return Valid;                                       };
private:
   void Set(int nChannel)                         { Channel = (Valid = (1 <= nChannel) && (nChannel <= NUM_ANALOGCHANNELS)) ? nChannel - 1 : 0; };
   BYTE Channel;
   bool Valid;
};


/**********************************************************************************************/

//                                        AnalogIO

/**********************************************************************************************/


union ADC_CONFIGURATION
{
   struct
   {
      BYTE gain               : 3;    // 000 = 0, 100 = 16;
      BYTE hold_excit : 1;    // hold excitation - if 0 excit voltage turns off after measure is complete
      BYTE channel    : 3;    // 000 - 101 for SingleEnded; 000,010,100 for Differential
      BYTE sd                 : 1;    // 1 = SingleEnded; 0 = Differntial

   };
   BYTE byte_value;
};

union ADC_EXCITATION
{
   struct
   {
      BYTE voltage    :4;     // valid values go from 0 - 10 decimal (0000 - 1010) and translate linearly to -5 to +5
      BYTE channel    :3;     // 000 - 101
      BYTE on_off             :1;     // 0 = off; 1 = on
   };
   BYTE byte_value;
};

struct CONFIG_BY_CHANNEL_STRUCT
{
   ADC_CONFIGURATION configuration;
   ADC_EXCITATION excitation;
   UINT16 filter_notch;
   UINT16 warm_up_delay;
   UINT16 period;
   UINT8  samples;
   BOOL poly_adjust;    // use to allow/prevent system from using the calibration polynomial
                        // the calibration polynomial can make low voltage readings less accurate.
};


class DECL AnalogIO_API : public IODevice
{
public:
   AnalogIO_API(BYTE AddrOfDevice = 0x04, BYTE AddrOfI2C = 0xFE);
   ~AnalogIO_API();

public:
   // high level exports - use these commands to get the appropriate reading from the analog module
   // calibration is automatic and may take up to 3.6 seconds

   I2CCODE SingleVoltageReading(const CAnalogChannel& channel, double& voltage_measurement)
   {return SingleReading(MEASUREVOLTAGE_OPCODE, channel, voltage_measurement);};
   I2CCODE MultipleVoltageReadings(const CAnalogChannel& channel, double* data_array, DWORD byte_size_of_data_array, int samples_to_get, int& samples_got, DWORD ms_timeout = INFINITE)
   {return MultipleReadings(MEASUREVOLTAGE_OPCODE, channel, data_array, byte_size_of_data_array, samples_to_get, samples_got, ms_timeout);};

   I2CCODE DoubleVoltageReading(const CAnalogChannel& voltage_channel, double& voltage_measurement, double& excitation_measurement);

   I2CCODE SingleCurrentReading(const CAnalogChannel& channel, double& current_measurement)
   {return SingleReading(MEASURECURRENT_OPCODE, channel, current_measurement);};
   I2CCODE MultipleCurrentReadings(const CAnalogChannel& channel, double* data_array, DWORD byte_size_of_data_array, int samples_to_get, int& samples_got, DWORD ms_timeout = INFINITE)
   {return MultipleReadings(MEASURECURRENT_OPCODE, channel, data_array, byte_size_of_data_array, samples_to_get, samples_got, ms_timeout);};

   I2CCODE SingleCurrent420maReading(const CAnalogChannel& channel, double& current_measurement)
   {return SingleReading(MEASURE420MA_OPCODE, channel, current_measurement);};
   I2CCODE MultipleCurrent420maReadings(const CAnalogChannel& channel, double* data_array, DWORD byte_size_of_data_array, int samples_to_get, int& samples_got, DWORD ms_timeout = INFINITE)
   {return MultipleReadings(MEASURE420MA_OPCODE, channel, data_array, byte_size_of_data_array, samples_to_get, samples_got, ms_timeout);};

   I2CCODE SingleResistanceDCReading(const CAnalogChannel& channel, double& resistancedc_measurement)
   {return SingleReading(MEASURERESISTANCEDC_OPCODE, channel, resistancedc_measurement);};
   I2CCODE MultipleResistanceDCReadings(const CAnalogChannel& channel, double* data_array, DWORD byte_size_of_data_array, int samples_to_get, int& samples_got, DWORD ms_timeout = INFINITE)
   {return MultipleReadings(MEASURERESISTANCEDC_OPCODE, channel, data_array, byte_size_of_data_array, samples_to_get, samples_got, ms_timeout);};

   I2CCODE SingleResistanceACReading(const CAnalogChannel& channel, double& resistanceac_measurement)
   {return SingleReading(MEASURERESISTANCEAC_OPCODE, channel, resistanceac_measurement);};
   I2CCODE MultipleResistanceACReadings(const CAnalogChannel& channel, double* data_array, DWORD byte_size_of_data_array, int samples_to_get, int& samples_got, DWORD ms_timeout = INFINITE)
   {return MultipleReadings(MEASURERESISTANCEAC_OPCODE, channel, data_array, byte_size_of_data_array, samples_to_get, samples_got, ms_timeout);};

   I2CCODE SingleThermistorReading(const CAnalogChannel& channel, double& thermistor_measurement)
   {return SingleReading(MEASURETHERMISTOR_OPCODE, channel, thermistor_measurement);};
   I2CCODE MultipleThermistorReadings(const CAnalogChannel& channel, double* data_array, DWORD byte_size_of_data_array, int samples_to_get, int& samples_got, DWORD ms_timeout = INFINITE)
   {return MultipleReadings(MEASURETHERMISTOR_OPCODE, channel, data_array, byte_size_of_data_array, samples_to_get, samples_got, ms_timeout);};

   I2CCODE RMYoungReading(UINT32& count, UINT32& time);

   I2CCODE ReadCalibrationData(void); // this is done automatically // this command will make sure that we have the calibration coefs from the module
   I2CCODE Calibrate(const CAnalogChannel& channel); // this is done automatically
public:
   // See analog io documentation for description of these configuration settings.
   void SetConfigurationGain(const CAnalogChannel& Channel, int Gain);
   void SetConfigurationSingleEnded(const CAnalogChannel& Channel);
   void SetConfigurationDifferential(const CAnalogChannel& Channel);
   
   // When excitation hold is on, excitation voltage stays on after measurement is complete.
   void SetConfigurationExcitationHoldOn(const CAnalogChannel& Channel);
   void SetConfigurationExcitationHoldOff(const CAnalogChannel& Channel);

   // Regarding excitation and channels: when reading on a specific Channel, a 
   // voltigate will be output on the ExcitationChannel all other references to 
   // Channel in regards to excitation should be the Channel on which the reading 
   // is being made, eg., to read on channel ReadCh and excite channel XCh with 5V, 
   // call SetExcitationChannel(ReadCh, XCh), SetExcitationVoltage(ReadCh, 5), and 
   // SetExcitationVoltageOn(ReadCh).
   void SetExcitationChannel(const CAnalogChannel& Channel, const CAnalogChannel& ExcitationChannel);
   void SetExcitationVoltage(const CAnalogChannel& Channel, int Voltage);
   void SetExcitationVoltageOn(const CAnalogChannel& Channel);
   void SetExcitationVoltageOff(const CAnalogChannel& Channel);
   
   void SetFilterNotch(const CAnalogChannel& Channel, UINT16 FilterNotch);
   void SetWarmUpDelay(const CAnalogChannel& Channel, UINT16 WarmUpDelay);
   void SetPeriod(const CAnalogChannel& Channel, int Period);     // valid period:  0 to 0xffff
   void SetSamples(const CAnalogChannel& Channel, int Samples);   // valid samples: 0 to 0xff
   void SetPolyAdjust(const CAnalogChannel& Channel, BOOL poly_adjust);

public:
   I2CCODE CmdSetAux1On()            { return CmdSetAux1(true);  };
   I2CCODE CmdSetAux1Off()           { return CmdSetAux1(false); };
   I2CCODE CmdSetAux1(bool High);  // sets the digital output aux1 to output a high/low signal
   I2CCODE CmdPulseOut(bool High, UINT16 PulseWidth_ms);   // causes aux1 to pulse high/low for a time period (and revert back to prev state)

public:
   const TCHAR * GetCmdStr(BYTE CmdOpCode);        // returns the string name of the cmd

public:
   // these are public so that the type of reading (current, voltage, etc) can be specified by a variable
   I2CCODE SingleReading(BYTE OpCode, const CAnalogChannel& channel, double& current_measurement);
   I2CCODE MultipleReadings(BYTE OpCode, const CAnalogChannel& channel, double * data_array, DWORD byte_size_of_data_array, int samples_to_get, int& samples_got, DWORD ms_timeout);

public:
   // Clear calibration information so Xpert will get it again.
   void InitCalInfo();

private:
   I2CCODE SendConfigCmd(I2CMSG& Msg, BYTE OpCode, CONFIG_BY_CHANNEL_STRUCT& config);
   I2CCODE CmdCalibrate(CONFIG_BY_CHANNEL_STRUCT& config);

private:
   void Notify(const I2CMSG& Msg);
   void Listen()   {Mgr.Listen(*this, DeviceAddr);};       // tells the i2cmrg to notify us whenever a message comes in for this device
   void UnListen() {Mgr.Unlisten(*this, DeviceAddr);};

private:
   // configuration handling
   CONFIG_BY_CHANNEL_STRUCT config_by_channel[16];
   CONFIG_BY_CHANNEL_STRUCT last_config_sent;
   void CopyAnalogConfig(CONFIG_BY_CHANNEL_STRUCT& dest, CONFIG_BY_CHANNEL_STRUCT& source);
   void DefaultConfig();

private:
   BOOL Read3ByteMeasurement(double& measurement, const I2CMSG& Msg, BYTE gain, BOOL poly_adjust);      // interprets voltage from a rxed msg
   BOOL Read6ByteMeasurement(double& voltage, double& current, const I2CMSG& Msg, BYTE gain, BOOL poly_adjust);
   BOOL DecodeMeasurement(const CAnalogChannel& channel, double& measurement, const I2CMSG& Msg);

   // events
   HANDLE measure_complete;        // used to notify a completed measurement (MultipleVoltageReading)
   HANDLE abort;                               // abort event - this might get passed in
private:
   // varibles used with MulitpleVoltageReading
   int             mvr_samples_requested;
   int             mvr_samples_collected;
   double* mvr_data_collected;
   CAnalogChannel mvr_current_channel;

private:
   // acquired from the analog io module using a ReadCalibrationData command
   double volt_coef_x1_pos;        // these adjust voltage readings to acquire 16 bit accuracy
   double volt_coef_x2_pos;
   double volt_coef_x0_pos;
   double volt_coef_x1_neg;
   double volt_coef_x2_neg;
   double volt_coef_x0_neg;   
   double volt_coefHI_x1_pos;        // these are the cal values for higain (x16)
   double volt_coefHI_x2_pos;
   double volt_coefHI_x0_pos;
   double volt_coefHI_x1_neg;
   double volt_coefHI_x2_neg;
   double volt_coefHI_x0_neg;
   double ohm20_error;             // used to adjust current/resist readings
   double ohm10k_error;    // used to adjust thermistor readings

   double AdjustForCalibrationData(double measurement, BYTE gain);
   bool   have_calibration_data;
   CTime TimeLastCal;

   double AdjustVoltageReading(double measurement, BYTE gain);
};


class DECL AnalogIO : public AnalogIO_API
{
public:
   AnalogIO(BYTE AddrOfDevice = 0x04, BYTE AddrOfI2C = 0xFE);
   ~AnalogIO();
   I2CCODE ReadResistance(const CAnalogChannel& Channel, 
         const CAnalogChannel& nExcitationVoltage, double& ReadData);
   I2CCODE ReadFrequency(int Period, bool TakeTwoReadings, double& Data);

private:
   UINT32 LastCount;
   UINT32 LastTime;
   UINT32 LastTick;
};

#endif // !defined(AFX_AnalogIO_H__65E6C3BC_87B9_11D3_8080_00105A1E5D53__INCLUDED_)
