#ifndef SSP_H
#define SSP_H

/*
   General format of a SSP Message:

   DLE SOH TO(n)/FROM(n) DLE STX FLAGSEQ(4) OPCODE LEN(2) DATA(n) DLE ETX CRC(2)

   Non-printable (binary) characters may appear in the DATA portion of the message, so
   in order to keep the 'DLE' tags identifiable, any DLE characters inside of the DATA
   portion are prefixed with a DLE. This makes it so a DLE ETX  can never be found anywhere
   but at the end of an SSP message.

   The CRC-16 is performed over all characters after the initial DLE SOH as this part of
   the message is always the same. The CRC is also performed on the 'raw' contents of the
   message (i.e. before DLE DLE sequences are converted to just DLE).

   The FLAGSEQ field is used to indentify unique packets so it needs to change when
   a new (non-retry) message is sent. It can be a simple incrementing counter or it could
   be the system tick count, etc.

   Although not indicated in the format description above, SSP supports multiple opcodes
   in a message. Additional opcodes simply follow with no seperator. No more opcodes can
   be detected by the fact that the opcode DLE is reserved. If you see an opcode of DLE
   then it means there are no more opcodes and you've parsed in to the DLE ETX indicating
   the end of the message.

   The actual opcodes, contents, and descriptions are beyond the scope of this quick summary.

   Please see the SSP Reference Manuals for further information.

*/

#include "Xpert/Win32Fix.h"

#undef DECL
#if defined(_WINDLL) || defined(XPERT_EXE)
#ifdef UTILS_DLL
#define DECL _declspec(dllexport)
#else
#define DECL _declspec(dllimport)
#endif
#else
#define DECL 
#endif


const UCHAR 
   DLE = 16,
   SOH = 1,
   STX = 2,
   ETX = 3,
   EOT = 4,
   ACK = 6,
   CTS = 6,   // { Used by software handshake }
   RTS = 17,  // { Used by software handshake }
   ESC = 27;

const
   SDLVersion = 'F';  // { Current SDL version }


struct TBertInfo
{
   DWORD GoodCount, BadCount, ErrorSet;
};

typedef TBertInfo* PBertInfo;

typedef char TTagName[10];  // Modula-2 string, won't have a NULL if it's exactly 10 chars

struct TTagDescriptor
{
   TTagName Name;       // { Name of this tag                    }
   UINT16 NameSize;     // { Size of the value names in bytes    }
   UINT16 StartSize;    // { Size of the start code in bytes     }
   UINT16 EvalSize;     // { Size of the eval code in bytes      }
   UINT16 StopSize;     // { Size of the stop code in bytes      }
   UINT16 ValueSize;    // { Size of the tag values in bytes     }
};
   
typedef enum TProtoOpcode
{
   OpAck,                // (* 00: Acknowledge an Opcode            *)
   OpNak,                // (* 01: Cannot perform an Opcode         *)
   OpCurdataReq,         // (* 02: Request for Current data         *)
   OpCurdata,            // (* 03: Current data report              *)

   OpTimeTagReq,         // (* 04: Request for Timed data           *)
   OpTimeTag,            // (* 05: Timed data report                *)
   OpTimeTagEnd,         // (* 06: End of time tag data             *)

   OpSelfTest,           // (* 07: Perform a selftest               *)
   OpClearStatus,        // (* 08: Clears system status             *)
   OpGetStatus,          // (* 09: Request system status            *)
   OpSystemStatus,       // (* 10: Here is the system status        *)

   OpStartSDL,           // (* 11: Start SDL system running         *)
   OpStopSDL,            // (* 12: Stop SDL system running          *)
   OpStartTag,           // (* 13: Runs the start code of a tag     *)
   OpEvalTag,            // (* 14: Runs the eval code of a tag      *)
   OpStopTag,            // (* 15: Runs the stop code of a tag      *)

   OpEOT,                // (* 16: Must be same as DLE. #16         *)

   OpSetClock,           // (* 17: This is the real time            *)
   OpReset,              // (* 18: Perform a reset                  *)
   OpEraseSetup,         // (* 19: Erases all setup information     *)

   OpSetupReq,           // (* 20: Ask for setup                    *)
   OpSetupInvalid,       // (* 21: Current setup is bad             *)
   OpSetupChanged,       // (* 22: Setup changed by terminal        *)

   OpTagInfo,            // (* 23: Single tag setup descriptor info *)
   OpTagData,            // (* 24: Setup data for this tag          *)
   OpTagEnd,             // (* 25: End of this tag                  *)
   OpSetupEnd,           // (* 26: End of the setup                 *)

   OpValueReq,           // (* 27: Request for tag data values      *)
   OpValue,              // (* 28: Here are tag data values         *)

   OpAlarm,              // (* 29: Alarm Exeception Report          *)

   OpSelectControl,      // (* 30: Select a controled point         *)
   OpBeginControl,       // (* 31: Begin control                    *)

   OpSWTableReady,       // (* 32: RTU Flow/Valvepos table is ready *)

   OpIllegal,            // (* 33: Illegal opcode rcvd              *)
   OpRestricted,         // (* 34: Operation is not avail to user   *)
   OpLogin,              // (* 35: Request to login                 *)
   OpBadPassword,        // (* 36: Login attempt failed             *)
   OpPleaseLogin,        // (* 37: User is logged out               *)
   OpLogout,             // (* 38: Log me out                       *)
   OpAllValueReq,        // (* 39: Request all values for a tag     *)
   OpAllValue,           // (* 40: Here are all the values          *)
   OpMail,               // (* 41: Here is a mail message           *)
   OpStopped,            // (* 42: SDL is stopped                   *)

   OpDDEPoke,            // (* 43: DDE Poke message                 *)
   OpDDERequest,         // (* 44: DDE Request message              *)
   OpDDEExecute,         // (* 45: DDE Execute message              *)
   OpDDEInitiate,        // (* 46: DDE Initiate message             *)
   OpDDETerminate,       // (* 47: DDE Terminate message            *)
   OpDDEAdvise,          // (* 48: DDE Advise message               *)
   OpDDEUnAdvise,        // (* 49: DDE UnAdvise message             *)

   OpBertStatusReq,      // (* 50: Request BERT status              *)
   OpBertStatus,         // (* 51: BERT status                      *)
   OpBertClear,          // (* 52: Clear BERT status                *)
   OpBertReq,            // (* 53: Request a BERT message           *)
   OpBert,               // (* 54: Dummy opcode since, non-SSP msg  *)
   OpAlert,              // (* 55: Alert over SSP message           *)

   // Special opcodes for XPert Remote Control Operations
   OpScreenReq,          // (* 56: Request a 320x240 2bpp RLE bmap  *)
   OpScreen,             // (* 57: The screen data itself           *)
   OpScreenDeltaReq,     // (* 58: Request just screen changes      *)    
   OpScreenDelta,        // (* 59: xor of the bmap w/previous bmap  *)  
   OpSendKey,            // (* 60: Send a keyevent structure        *)
   OpSendMouse,          // (* 61: Send a mouseevent structure      *) 

   // Socket over SSP Support Opcodes
   OpSocketOpen,         // (* 62: Open a connection to a server    *)
   OpSocketOpenAck,      // (* 63: Response to a socket open req    *)
   OpSocketOpenNak,      // (* 64: Response to a socket open req    *)
   OpSocketWrite,        // (* 65: Close a connection               *)
   OpSocketWriteAck,     // (* 66: Close a connection               *)
   OpSocketWriteNak,     // (* 67: Close a connection               *)
   OpSocketClose,        // (* 68: Close a connection               *)

   OpFileReq,            // (* 69: Request a file                   *)
   OpFileStart,          // (* 70: Start of a file transfer         *)
   OpFileData,           // (* 71: File data                        *)
   OpFileEnd,            // (* 72: End of transfer                  *)

   OpDir,                // (* 73: Perform a directory              *)
   OpPing,               // (* 74: Simple command to check connect  *)
   OpDeleteFile,         // (* 75: Delete a file                    *)
   OpRemoveDir,          // (* 76: Remove a Dir                     *)
   OpRun,                // (* 77: Run program                      *)
   OpMakeDir,            // (* 78: Make Directory                   *)
   OpRename,             // (* 79: Rename file                      *)       

   OpPacket,             // (* 80: Raw SSP Packet                   *)
   OpData,               // (* 81: Raw SSP Data (no to/from/etc)    *)

   // Opcodes for Com Port redirection in Remote
   OpComOpen,            // (* 82: Attach to a Com Port                 *)
   OpComClose,           // (* 83: End attachment                       *) 
   OpComLock,            // (* 84: Gain exclusive access to port        *)
   OpComUnLock,          // (* 85: Release exclusive access             *)
   OpComData,            // (* 86: Byte data to/from the port           *)
   OpComDataWait,        // (* 87: OpComData with ack when data sent    *)
   OpComPacket,          // (* 88: SSP Packet data to/from the port     *)
   OpComPacketWait,      // (* 89: OpComPacket with ack when data sent  *)
   OpComControl,         // (* 90: Set Control Line Signals             *)
   OpComGetStatus,       // (* 91: Get Control Line Status              *)
   OpComStatus,          // (* 92: Control Line Status                  *)
   OpComCapture,         // (* 93: Enable/Disable Capture of Data       *) 
   OpComGetPortList,     // (* 94: Request port list                    *) 
   OpComPortList,        // (* 95: List of ports controlled             *) 
   OpComRunHook,         // (* 96: RemoteOp cmd to run special code     *) 
   OpComGetOptions,      // (* 97: Get various options and flags        *)
   OpComOptions,         // (* 98: Packet returned by OpComOptions      *)
   OpComSetOptions,      // (* 99: Set/Clear various port options       *)
   OpComGetCommState,    // (*100: Get the Comm Port's DCB              *)
   OpComCommState,       // (*101: Packet returned by OpComGetCommState *)
   OpComSetCommState,    // (*102: Set the Comm Port's DCB              *)
   OpComSetExtendedCmds, // (*103: Set list of extended commands        *)
   OpComRunCommand,      // (*104: Extended command to run              *)

   // Opcodes used by XPert and Remote
   OpGetSSPStat,         // (*105: Retrieve SSP statistics for a port   *)
   OpSSPStat,            // (*106: SSP stats from a port                *)
   OpClearSSPStat,       // (*107: Clear SSP Statistics for a port      *)
   OpDebugMessage,       // (*108: ASCII Debug message                  *)
   OpSetDebugFilter,     // (*109: Sets the debug message filter level  *)
   OpSendMail,           // (*111: Send a mail message to a remote port *)
   OpGetMailBox,         // (*111: Retrieve a mail box from a port      *)
   OpMailBox,            // (*112: String of messages from a mailbox    *)
   OpGetVersionInfo,     // (*113: Get version information              *)
   OpVersionInfo,        // (*114: String with version information      *)
   OpStartWatchDog,      // (*115: Start Remote's watchdog              *)
   OpStopWatchDog,       // (*116: Stop Remote's watchdog               *)
   OpCheckWatchDog,      // (*117: Check if XPert is running ok         *)

   // More Opcodes for Com Port redirection in Remote
   OpComRequest,         // (*118: Send an SSP packet, and wait for reply       *)
   OpComWaitMessage,     // (*119: Wait for reply                               *)
   OpComReply,           // (*120: Reply packet for a request or a wait message *)

   // Misc
   OpSaveSetup,           // (*121: Forces a save setup to occur         *)

   // New and Improved XPert Remote Control Operations with better compression
   OpScreenDetect,       // (* 122: Request 320x240 2bpp RLE bmap (event driven) *)
   OpScreen2,            // (* 123: Full screen data                             *)
   OpScreenDelta2,       // (* 124: xor of the bmap w/previous bmap              *)  

   // More Opcodes used by XPert and Remote
   OpComRestart,         // (* 125: Restart Remote, Update autoexec              *) 

   // Misc
   OpSetPacketSize,      // (* 126: Change SSP Packetsize for a session          *)
   OpGetAccessLevel,     // (* 127: Request current access level                 *)
   OpAccessLevel,        // (* 128: Return current access level                  *)
   OpSpecialCmd,         // (* 129: Execute a special command                    *)
   OpSpecialData,        // (* 130: Results of a special command                 *)
   OpSpecialDataEnd,     // (* 131: End of a special data                        *)

   OpShutdown,           // (* 132: Shutdown the main app                        *)
   OpStartup             // (* 133: Startup the main app                         *)

};

typedef enum TDataType
{
   dt_boolean, dt_char, dt_cardinal, dt_integer, dt_long, dt_address, dt_sched, dt_time,
   dt_real, dt_string, dt_pointer, dt_alarm, dt_nil, dt_logptr, dt_longstr
};

// Enumeration used by OpComSetOptions and OpComOptions packet
typedef enum TComOptions
{
   IGNORECD=1<<0, LOGGINGIN=1<<1, LOGGEDIN=1<<2, ISMODEM=1<<3, ISRADIO=1<<4, ISDIRECT=1<<5,
   ISVOICE=1<<6, DEBUGCOMMANDS=1<<7, QUICKPROTOCOL=1<<8, DIALIN=1<<9, SSPLOGIN=1<<10, 
   ENABLECOMMANDPROMPT=1<<11, ENABLESSP=1<<12, ENABLECOMMANDPARSER=1<<13, ISSSPONLY=1<<14,
   ISRS485=1<<15, ENABLESSPPARSER=1<<16
};

extern "C" DECL int MaxSSPPacketSize;
extern "C" DECL int DesiredSSPPacketSize;
extern "C" DECL int SSPSafetyMargin;

extern "C" DECL DWORD GetFlagSeq();

extern "C" DECL WORD DoCRC16(LPCSTR Data, int Len);

struct TDataInfo
{
	long nRefs;        // reference count
	int nDataLength;
	int nAllocLength;
	// char data[nAllocLength]
	char* Data()	  { return (char *)(this+1); }
};


/*
   TData is a character array class which supports reference counting and dynamic sizeing.
   Reference counting allows copies of the array to be made without actually copying the data.
   When you want to write to the area, the references must be explicitly broken.
   This can be done by using SetSize or SetLength. The += and [] operators also create
   a new copy as needed. Once any references are broken you can directly access the array
   from 0 to GetSize()-1 bytes using the () operator. 

   Here's some examples of TData usage:

   TData a;
   TData b;
   TData c;
   a.SetSize(100);     // Make sure there is room for at least 100 bytes
   *a(0) = 'H';        // () returns a writeable char pointer
   *a(1) = 'E';
   *a(2) = 'L';
   *a(3) = 'L';
   *a(4) = 'O';
   a.SetLength(5);     // Set the length (it isn't implied by nulls cause there are no null bytes)
   a += '!';           // Concatenation automatically grows and sets the length
   b = a;              // This is a smart copy (no data is actually moved)
   b[0] = 'B';         // The [] operator will make a copy if necessary
   c = a;
   c.SetSize(20);      // SetSize() and SetLength() will also force a new copy to be made
   memset(c(), 0, 20); // Fill the string up with nulls ...
*/

class DECL TData
{
public:
   TData();
   TData(const TData& Data)              {Init(); *this=Data;                    };
   TData(const char* Buffer)             {Init(); *this=Buffer;                  };
   TData(char Ch)                        {Init(); *this=Ch;                      };
   ~TData();
   const TData& operator=(const TData& Data);
   TData& operator=(const char* Buffer);
   TData& operator=(char Ch);
   void SetData(const char* Buffer, int iLen);
   void operator+=(char Ch); 
   TData operator+(const TData& Data) const; 
   void operator+=(const TData& Data); 
   TData operator+(const char* Buffer) const;
   void operator+=(const char* Buffer); 
   void AddData(const char* Buffer, int iLen);
   char& operator[](int Pos)             {SetSize(Pos); return pData[Pos];       };
   char operator[](int Pos) const;
   LPSTR operator()()                    {return pData;                          };
   LPSTR operator()(int Pos)             {return pData+Pos;                      };
   operator LPSTR() const                {return (LPSTR) pData;                  };
   operator LPCSTR() const               {return (LPCSTR) pData;                 };
   operator LPCVOID() const              {return (LPCVOID) pData;                };
   operator CString() const;
   int GetLength() const                 {return GetInfo()->nDataLength;         };
   int GetSize() const                   {return GetInfo()->nAllocLength;        };
   void SetSize(int MinLength);
   void SetLength(int NewLength);
   bool IsEmpty() const                  {return GetLength() == 0;               };
   void Empty();
   void Clean()                          {Release();                             };
   int Delete(int iIndex, int iCount = 1);
   int Find(char ch, int iStart = 0) const;
private:
   void Init();
   void AllocBuffer(int nLen);
   void Release(TDataInfo* pInfo);
   void Release();
   void GrowBuffer(int nLen);
   TDataInfo* GetInfo() const	{ return ((TDataInfo *) pData)-1; };
protected:
   char* pData;
};


// A class to build and parse the opcode and data portion of an SSP Message
class DECL TSSPData
{
public:
   TSSPData();
   TSSPData(const TSSPData& Data);
   virtual ~TSSPData();
   TSSPData& operator=(const TSSPData& d);
//
   TProtoOpcode GetFirstOpcode();
   int GetFirstLen();
   LPCSTR GetFirstData();
//
   TProtoOpcode GetNextOpcode();
   int GetNextLen();
   LPCSTR GetNextData();
//
   bool Read(TProtoOpcode& Opcode, int& DataLen);
   bool Read(void* Data, int DataLen);
   bool Read(char* Data);
   bool Read(CString& Data);
   bool Read(BYTE& Data)                 {return Read(&Data, sizeof(BYTE));};
   bool Read(char& Data)                 {return Read(&Data, sizeof(char));};
   bool Read(WORD& Data)                 {return Read(&Data, sizeof(WORD));};
   bool Read(DWORD& Data)                {return Read(&Data, sizeof(DWORD));};
   bool Read(int& Data)                  {return Read(&Data, sizeof(int));};
   bool Read(double& Data);
//
   bool Write(TProtoOpcode Opcode);
   bool Write(const char* Data);
   bool Write(double Data);
#ifdef UNICODE
   bool Write(LPCTSTR Data);
#endif
   bool Write(const void* Data, int DataLen);
   bool Write(BYTE Data)                 {return Write(&Data, sizeof(BYTE));};
   bool Write(char Data)                 {return Write(&Data, sizeof(char));};
   bool Write(WORD Data)                 {return Write(&Data, sizeof(WORD));};
   bool Write(DWORD Data)                {return Write(&Data, sizeof(DWORD));};
   bool Write(int Data)                  {return Write(&Data, sizeof(int));};
   bool Write(TSSPData& Data);     // Encapulate Data
   void Rewind();
   void Clear();
   void StartEncapsulate();              // Encapsulate a data packet in a OpData packet
   void StopEncapsulate();               // Update the length of the OpData packet
//
   const TData& Get();
   void Set(const TData& Data);
   LPCSTR GetCharPtr(int DataLen);
// SSP Helper routines
   double LongR(unsigned long A);
   unsigned long ShortR(double R);
private:
   friend class TSSPPacket;
   TData Buffer;
   int Pos;
   int OpcodePos;
   int NextOpcodePos;
};

// A class to build and parse entire SSP messages
class DECL TSSPPacket
{
public:
   TSSPPacket();
   virtual ~TSSPPacket();
//
   bool Write(const CString& To, const CString& From, UINT32 FlagSeq, TSSPData& SSPData);
   bool Read(CString& To, CString& From, UINT32& FlagSeq, TSSPData& SSPData);
//
   const TData& Get()                    {return Buffer;  };
   void Set(const TData& Data)           {Buffer=Data;    };

// SSP Helpfull functions
   static CString GetUnitID(CString ToPath);
   static CString GetRepeatToPath(CString ToPath);
   static int GetRepeatToPort(CString ToPath);
   static CString GetRepeatFromPath(CString ToPath, CString FromPath, CString Port);
   static bool MyMessage(CString ToPath, CString UnitID);
   static bool RepeatMessage(CString ToPath, CString UnitID);

public: // Validation helper functions
   static bool IsValid(const TData& Data);
   static bool IsMine(const TData& Data, CString UnitID);

private:
   int WriteDLE(LPSTR Buffer, LPCSTR Data, int Len);
   int ReadDLEFlagSeq(LPSTR Buffer, LPCSTR Data, int BufLen);
   int ReadDLEData(LPSTR Buffer, LPCSTR Data, int DataLen);

private:   
   TData Buffer;
};

#endif // SSP_H
